import json, os, re
from functools import lru_cache
from typing import Callable, Literal, Optional, Union

import matplotlib.pyplot as plt, numpy as np
from joblib import Parallel, delayed, dump

from .._utils import (
    _TimeseriesExtractorGetter,
    _check_kwargs,
    _check_confound_names,
    _check_parcel_approach,
    _extract_timeseries,
    _logger,
)

LG = _logger(__name__)


# Custom error class
class BIDSQueryError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return f"{self.message}"


class TimeseriesExtractor(_TimeseriesExtractorGetter):
    """
    Timeseries Extractor Class.

    Initializes the Timeseries Extractor class.

    Parameters
    ----------
    space: :obj:`str`, default="MNI152NLin2009cAsym"
        The standard template space that the preprocessed bold data is registered to. Used for querying with pybids
        to locate preprocessed BOLD-related files.

    parcel_approach: :obj:`dict[str, dict[str, Union[str, int]]]` or :obj:`os.PathLike`, \
                     default={"Schaefer": {"n_rois": 400, "yeo_networks": 7, "resolution_mm": 1}}
        The approach to parcellate NifTI images. This must be a nested dictionary with the first key being the
        parcellation name. Currently, only "Schaefer", "AAL", and "Custom" are supported. Recognized second level
        keys (sub-keys) are listed below:

        - For "Schaefer":

            - "n_rois": The number of ROIs (100, 200, 300, 400, 500, 600, 700, 800, 900, or 1000). Defaults to 400.
            - "yeo_networks": The number of Yeo networks (7 or 17). Defaults to 7.
            - "resolution_mm": The spatial resolution of the parcellation in millimeters (1 or 2). Defaults to 1.

        - For "AAL":

            - "version": The version of the AAL atlas used ("SPM5", "SPM8", "SPM12", or "3v2"). Defaults to "SPM12" if ``{"AAL": {}}`` is supplied.

        - For "Custom":

            - "maps": Directory path to the location of the parcellation file.
            - "nodes": A list of node names in the order of the label IDs in the parcellation.
            - "regions": The regions or networks in the parcellation.

        Refer to `Nilearn's Fetch Schaefer Documentation
        <https://nilearn.github.io/stable/modules/generated/nilearn.datasets.fetch_atlas_schaefer_2018.html#nilearn.datasets.fetch_atlas_schaefer_2018>`_
        and `Nilearn's Fetch AAL Documentation
        <https://nilearn.github.io/stable/modules/generated/nilearn.datasets.fetch_atlas_aal.html#nilearn.datasets.fetch_atlas_aal>`_
        for more information about the "Schaefer" and "AAL" sub-keys. Also, refer to the "Note" section below for an
        explanation of the "Custom" sub-keys.

    standardize: {"zscore_sample", "zscore", "psc", True, False}, default="zscore_sample"
        Standardizes the timeseries. Refer to ``nilearn.maskers.NiftiLabelsMasker`` for an explanation of each
        available option.

    detrend: :obj:`bool`, default=True
        Detrends the timeseries during extraction.

    low_pass: :obj:`float`, :obj:`int`, or :obj:`None`, default=None
        Filters out signals above the specified cutoff frequency.

    high_pass: :obj:`float`, :obj:`int`, or :obj:`None``, default=None
        Filters out signals below the specified cutoff frequency.

    fwhm: :obj:`float`, :obj:`int`, or :obj:`None`, default=None
        Applies spatial smoothing to data (in millimeters). Note that using a parcellation already averages voxels
        within parcel boundaries, which can improve signal-to-noise ratio (SNR) assuming Gaussian noise
        distribution. However, smoothing may also blur parcel boundaries.

    use_confounds: :obj:`bool`, default=True
        Perform nuisance regression using the default or user-specified confounds in ``confound_names`` when
        extracting timeseries. Note, the confound tsv files must be located in the same directory as the preprocessed
        BOLD images.

    confound_names: :obj:`list[str]` or :obj:`None`, default=None
        The names of the confounds to extract from the confound tsv files. If None, default confounds are used, which
        consists of all cosine-basis parameters, the six-head motion parameters and their first-order derivatives,
        and the first six combined acompcor components. Additionally, the names of these confounds follow the
        naming scheme of confounds in fMRIPrep versions ``>= 1.2.0``. Note, an asterisk ("*") can be used to find
        confound names that start with the term preceding the asterisk. For instance, "cosine*" will find all confound
        names in the confound files starting with "cosine".

    fd_threshold: :obj:`float`, :obj:`dict[str, float]`, or :obj:`None`, default=None
        Sets a threshold for removing volumes that exceed a specified level. This parameter applies the condition
        ``framewise displacement (FD) > threshold``, meaning that only frames with a framewise displacement above the
        set threshold are removed. This requires a column named `framewise_displacement` in the
        confounds file and ``use_confounds`` set to True. Additionally, `framewise_displacement` should be
        specified in ``confound_names`` if using this parameter. By default, censoring is done after nuisance
        regression; however, this behavior can be modified with the "use_sample_mask" key to censor prior to
        nuisance regression. If, ``fd_threshold`` is a dictionary, the following keys can be specified:

        - "threshold": A float or integer value. Volumes with a `framewise_displacement` value exceeding this threshold
          are removed.
        - "outlier_percentage": A float value between 0 and 1 representing a percentage. Runs where the proportion of
          volumes exceeding the "threshold" is higher than this percentage are removed. If ``condition`` is specified
          in ``self.get_bold``, only the runs where the proportion of volumes exceeds this value for the specific
          condition of interest are removed. Note, this proportion is calculated after dummy scans have been removed.
          A warning is issued whenever a run is flagged.
        - "n_before": An integer value indicating the number of volumes to scrub before the flagged volume. Hence,
          if frame 5 is flagged and "n_before" is 2, then volumes 3, 4, and 5 are scrubbed.
        - "n_after": An integer indicating the number of volumes to scrub after to the flagged volume. Hence,
          if frame 5 is flagged and "n_after" is 2, then volumes 5, 6, and 7 are scrubbed.
        - "use_sample_mask": A boolean value. If True, a sample mask is generated and passed to the ``sample_mask``
          parameter in Nilearn's ``NiftiLabelsMasker`` to censor prior to nuisance regression. Internally,
          ``clean__extrapolate`` is set to False and passed to ``NiftiLabelsMasker``, which prevents censored
          volumes at the end from being interpolated prior to applying the butterworth filter. See
          `Nilearn's Signal Clean Documentation
          <https://nilearn.github.io/stable/modules/generated/nilearn.signal.clean.html>`_ and
          `Nilearn's NiftiLabelsMasker Documentation
          <https://nilearn.github.io/stable/modules/generated/nilearn.maskers.NiftiLabelsMasker.html>`_ for how Nilearn
          handles censored volumes when ``sample_mask`` is used.  If this key is set to False, data is only censored
          after nuisance regression, which is the default behavior.

        **Note**: If "use_sample_mask" is not True and the ``standardize`` is not False, it is recommended to apply an
        additional within-run standardization (using ``neurocaps.analysis.standardize``) once outlier volumes have been
        removed.

        .. versionadded:: 0.18.8 "use_sample_mask"

    n_acompcor_separate: :obj:`int` or :obj:`None`, default=None
        Specifies the number of separate acompcor components derived from white-matter (WM) and cerebrospinal
        fluid (CSF) masks to use. For example, if set to 5, the first five components from the WM mask
        and the first five from the CSF mask will be used, totaling ten acompcor components. If this parameter is
        not None, any acompcor components listed in ``confound_names`` will be disregarded. To use acompcor
        components derived from combined masks (WM & CSF), leave this parameter as None and list the specific
        acompcors of interest in ``confound_names``.

    dummy_scans: :obj:`int`, :obj:`dict[str, Union[bool, int]]`, or :obj:`None`, default=None
        Removes the first `n` volumes before extracting the timeseries. If, ``dummy_scans`` is a dictionary,
        the following keys can be used:

        - "auto": A boolean value. If True, the number of dummy scans removed depend on the number of
          "non_steady_state_outlier_XX" columns in the participants fMRIPrep confounds tsv file. For instance, if
          there are two "non_steady_state_outlier_XX" columns detected, then ``dummy_scans`` is set to two since
          there is one "non_steady_state_outlier_XX" per outlier volume for fMRIPrep. This is assessed for each run of
          all participants so ``dummy_scans`` depends on the number number of "non_steady_state_outlier_XX" in the
          confound file associated with the specific participant, task, and run number.
        - "min": An integer value indicating the minimum dummy scans to discard. The "auto" sub-key must be True
          for this to work. If, for instance, only two "non_steady_state_outlier_XX" columns are detected but the
          "min" is set to three, then three dummy volumes will be discarded.
        - "max": An integer value indicating the maximum dummy scans to discard. The "auto" sub-key must be True
          for this to work. If, for instance, six "non_steady_state_outlier_XX" columns are detected but the
          "max" is set to five, then five dummy volumes will be discarded.

    dtype: :obj:`str` or "auto", default=None
        The NumPy dtype the NIfTI images are converted to when passed to Nilearn's ``load_img`` function.


    Properties
    ----------
    space: str
        The standard template space that the preprocessed BOLD data is registered to. The space can also be set after
        class initialization using ``self.space = "New Space"`` if the template space needs to be changed.

    parcel_approach: :obj:`dict[str, dict[str, Union[os.PathLike, list[str]]]]` or \
                     :obj:`dict[str, dict[str, Union[os.PathLike, list[str], dict[str, dict[str, list[int]]]]]]`
        A dictionary containing information about the parcellation. Can also be used as a setter, which accepts a
        dictionary or a dictionary saved as pickle file. If "Schaefer" or "AAL" was specified during
        initialization of the ``TimeseriesExtractor`` class, then ``nilearn.datasets.fetch_atlas_schaefer_2018``
        and ``nilearn.datasets.fetch_atlas_aal`` will be used to obtain the "maps" and the "nodes". Then string
        splitting is used on the "nodes" to obtain the "regions":

        ::

            # Structure of Schaefer
            {
                "Schaefer":
                {
                    "maps": "path/to/parcellation.nii.gz",
                    "nodes": ["LH_Vis1", "LH_SomSot1", "RH_Vis1", "RH_Somsot1"],
                    "regions": ["Vis", "SomSot"]
                }
            }

            # Structure of AAL
            {
                "AAL":
                {
                    "maps": "path/to/parcellation.nii.gz",
                    "nodes": ["Precentral_L", "Precentral_R", "Frontal_Sup_L", "Frontal_Sup_R"],
                    "regions": ["Precentral", "Frontal"]
                }
            }

        Refer to the example for "Custom" in the Note section below for the expected structure.

    signal_clean_info: :obj:`dict[str, Union[bool, int, float, str]]` or :obj:`None`
        Dictionary containing parameters for signal cleaning specified during initialization of the
        ``TimeseriesExtractor`` class. This information includes ``standardize``, ``detrend``, ``low_pass``,
        ``high_pass``, ``fwhm``, ``dummy_scans``, ``use_confounds``, ``n_compcor_separate``, and ``fd_threshold``.

    task_info: :obj:`dict[str, Union[str, int]]` or :obj:`None`
        If ``self.get_bold()`` ran, is a dictionary containing all task-related information such as ``task``,
        ``condition``, ``session``, ``runs``, and ``tr`` (if specified) else None.

    subject_ids: :obj:`list[str]` or :obj:`None`
        A list containing all subject IDs that have retrieved from pybids and subjected to timeseries
        extraction.

    n_cores: :obj:`int` or :obj:`None`
        Number of cores used for multiprocessing with Joblib.

    subject_timeseries: :obj:`dict[str, dict[str, np.ndarray]]` or :obj:`None`
        A dictionary mapping subject IDs to their run IDs and their associated timeseries (TRs x ROIs) as a NumPy array.
        Can also be a path to a pickle file containing this same structure. If this property needs to be deleted due
        to memory issues,  ``delattr(self, "_subject_timeseries")`` (version < 0.18.10) or
        ``del self.subject_timeseries`` (version >= 0.18.10) can be used to delete this property and only have it
        return None. The structure is as follows:

        ::

            subject_timeseries = {
                    "101": {
                        "run-0": np.array([...]), # Shape: TRs x ROIs
                        "run-1": np.array([...]), # Shape: TRs x ROIs
                        "run-2": np.array([...]), # Shape: TRs x ROIs
                    },
                    "102": {
                        "run-0": np.array([...]), # Shape: TRs x ROIs
                        "run-1": np.array([...]), # Shape: TRs x ROIs
                    }
                }

    Note
    ----
    **Passed Parameters**: ``standardize``, ``detrend``, ``low_pass``, ``high_pass``, ``fwhm``, and nuisance
    regression (``confound_names``) uses ``nilearn.maskers.NiftiLabelsMasker``. The ``dtype`` parameter is used by
    ``nilearn.image.load_img``. For framewise displacement, if the "use_sample_mask" key is set to True in the
    ``fd_threshold`` dictionary, then a boolean sample mask is generated (setting indices corresponding to high motion
    volumes as False) and is passed to the ``sample_mask`` parameter in ``nilearn.maskers.NiftiLabelsMasker``.

    **Custom Parcellations**: If using a "Custom" parcellation approach, ensure that the parcellation is
    lateralized (where each region/network has nodes in the left and right hemisphere). This is due to certain
    visualization functions assuming that each region consists of left and right hemisphere nodes. Additionally,
    certain visualization functions in this class also assume that the background label is 0. Therefore, do not add a
    background label in the "nodes" or "regions" keys.

    The recognized sub-keys for the "Custom" parcellation approach includes:

    - "maps": Directory path containing the parcellation file in a supported format (e.g., .nii or .nii.gz for NifTI).
    - "nodes": A list of all node labels. The node labels should be arranged in ascending order based on their
      numerical IDs from the parcellation files. The node with the lowest numerical label in the parcellation file
      should occupy the 0th index in the list, regardless of its actual numerical value. For instance, if the numerical
      IDs are sequential, and the lowest, non-background numerical ID in the parcellation is "1" which corresponds
      to "left hemisphere visual cortex area" ("LH_Vis1"), then "LH_Vis1" should occupy the 0th element in this list.
      Even if the numerical IDs are non-sequential and the earliest non-background, numerical ID is "2000"
      (assuming "0" is the background), then the node label corresponding to "2000" should occupy the 0th element of
      this list.

      ::

            # Example of numerical label IDs and their organization in the "nodes" key
            "nodes": {
                "LH_Vis1",          # Corresponds to parcellation label 2000; lowest non-background numerical ID
                "LH_Vis2",          # Corresponds to parcellation label 2100; second lowest non-background numerical ID
                "LH_Hippocampus",   # Corresponds to parcellation label 2150; third lowest non-background numerical ID
                "RH_Vis1",          # Corresponds to parcellation label 2200; fourth lowest non-background numerical ID
                "RH_Vis2",          # Corresponds to parcellation label 2220; fifth lowest non-background numerical ID
                "RH_Hippocampus"    # Corresponds to parcellation label 2300; sixth lowest non-background numerical ID
            }

    - "regions": A dictionary defining major brain regions or networks. Each region should list node indices under
      "lh" (left hemisphere) and "rh" (right hemisphere) to specify the respective nodes. Both the "lh" and "rh"
      sub-keys should contain the indices of the nodes belonging to each region/hemisphere pair, as determined
      by the order/index in the "nodes" list. The naming of the sub-keys defining the major brain regions or networks
      have zero naming requirements and simply define the nodes belonging to the same name.

      ::

            # Example of the "regions" sub-keys
            "regions": {
                "Visual": {
                    "lh": [0, 1], # Corresponds to "LH_Vis1" and "LH_Vis2"
                    "rh": [3, 4]  # Corresponds to "RH_Vis1" and "RH_Vis2"
                },
                "Hippocampus": {
                    "lh": [2], # Corresponds to "LH_Hippocampus"
                    "rh": [5]  # Corresponds to "RH_Hippocampus"
                }
            }

    The provided example demonstrates setting up a custom parcellation containing nodes for the visual network (Vis)
    and hippocampus regions in full:

    ::

        parcel_approach = {
            "Custom": {
                "maps": "/location/to/parcellation.nii.gz",
                "nodes": [
                    "LH_Vis1",
                    "LH_Vis2",
                    "LH_Hippocampus",
                    "RH_Vis1",
                    "RH_Vis2",
                    "RH_Hippocampus"
                ],
                "regions": {
                    "Visual": {
                        "lh": [0, 1],
                        "rh": [3, 4]
                    },
                    "Hippocampus": {
                        "lh": [2],
                        "rh": [5]
                    }
                }
            }
        }

    **Note**: Different sub-keys are required depending on the function used. Refer to the Note section under each
    function for information regarding the sub-keys required for that specific function.
    """

    def __init__(
        self,
        space: str = "MNI152NLin2009cAsym",
        parcel_approach: Union[dict[str, dict[str, Union[str, int]]], os.PathLike] = {
            "Schaefer": {"n_rois": 400, "yeo_networks": 7, "resolution_mm": 1}
        },
        standardize: Union[bool, Literal["zscore_sample", "zscore", "psc"]] = "zscore_sample",
        detrend: bool = True,
        low_pass: Optional[Union[float, int]] = None,
        high_pass: Optional[Union[float, int]] = None,
        fwhm: Optional[Union[float, int]] = None,
        use_confounds: bool = True,
        confound_names: Optional[list[str]] = None,
        fd_threshold: Optional[Union[float, dict[str, Union[bool, float, int]]]] = None,
        n_acompcor_separate: Optional[int] = None,
        dummy_scans: Optional[Union[int, dict[str, Union[bool, int]]]] = None,
        dtype: Union[str, Literal["auto"]] = None,
    ) -> None:

        self._space = space

        if use_confounds is False and fd_threshold is not None:
            LG.warning(
                "`fd_threshold` specified but `use_confounds` is not True so removal of volumes after "
                "nuisance regression will not be done since the fMRIPrep confounds tsv file is needed."
            )
        elif use_confounds is True and fd_threshold:
            self._validate_init_params("fd_threshold", fd_threshold)

        # Check dummy_scans and fd_threshold
        if dummy_scans:
            self._validate_init_params("dummy_scans", dummy_scans)

        self._signal_clean_info = {
            "masker_init": {
                "standardize": standardize,
                "detrend": detrend,
                "low_pass": low_pass,
                "high_pass": high_pass,
                "smoothing_fwhm": fwhm,
            },
            "dummy_scans": dummy_scans,
            "n_acompcor_separate": n_acompcor_separate,
            "fd_threshold": None,
            "use_confounds": use_confounds,
            "dtype": dtype,
        }
        # Check parcel_approach
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_approach)

        if self._signal_clean_info["use_confounds"]:
            self._signal_clean_info["confound_names"] = _check_confound_names(
                high_pass, confound_names, n_acompcor_separate
            )

            self._signal_clean_info["fd_threshold"] = fd_threshold

    @staticmethod
    def _validate_init_params(param, struct):

        mandatory_keys = {"dummy_scans": {"auto": bool}, "fd_threshold": {"threshold": (float, int)}}

        optional_keys = {
            "dummy_scans": {"min": int, "max": int},
            "fd_threshold": {"n_before": int, "n_after": int, "outlier_percentage": float, "use_sample_mask": bool},
        }

        valid_types = (dict, int) if param == "dummy_scans" else (dict, float, int)
        error_msg = "dictionary or integer" if param == "dummy_scans" else "dictionary, float, or integer"

        if not isinstance(struct, valid_types):
            raise TypeError(f"`{param}` must be a {error_msg}.")

        if isinstance(struct, dict):
            # Check mandatory keys
            key = list(mandatory_keys[param].keys())[0]
            if key not in struct:
                raise KeyError(f"'{key}' is a mandatory key when `{param}` is a dictionary.")

            if not isinstance(struct[key], mandatory_keys[param][key]):
                raise TypeError(f"'{key}' must be a {'boolean' if key == 'auto' else 'float or integer'}.")

            # Check optional keys
            for key in optional_keys[param]:
                if key in struct:
                    if not isinstance(struct[key], optional_keys[param][key]):
                        type_msg = (
                            "a boolean"
                            if key == "use_sample_mask"
                            else ("a float" if key == "outlier_percentage" else "an integer")
                        )
                        raise TypeError(f"'{key}' must be {type_msg}.")

                    # Additional check for "outlier_percentage"
                    if key == "outlier_percentage" and not 0 < struct[key] < 1:
                        raise ValueError("'outlier_percentage' must be float between 0 and 1.")

            # Check invalid keys
            set_diff = set(struct) - set(optional_keys[param]) - set(mandatory_keys[param])

            if set_diff:
                formatted_string = ", ".join(["'{a}'".format(a=x) for x in set_diff])
                LG.warning(
                    f"The following invalid keys have been found in `{param}` and will be ignored: {formatted_string}."
                )

    def get_bold(
        self,
        bids_dir: os.PathLike,
        task: str,
        session: Optional[Union[int, str]] = None,
        runs: Optional[Union[int, str, list[int], list[str]]] = None,
        condition: Optional[str] = None,
        condition_tr_shift: int = 0,
        tr: Optional[Union[int, float]] = None,
        slice_time_ref: Union[int, float] = 0.0,
        run_subjects: Optional[list[str]] = None,
        exclude_subjects: Optional[list[str]] = None,
        exclude_niftis: Optional[list[str]] = None,
        pipeline_name: Optional[str] = None,
        n_cores: Optional[int] = None,
        parallel_log_config: Optional[dict[str, Union[Callable, int]]] = None,
        verbose: bool = True,
        flush: bool = False,
    ) -> None:
        """
        Retrieve Preprocessed BOLD Data from BIDS Datasets.

        This function uses pybids for querying and requires the BOLD data directory (specified in ``bids_dir``) to be
        BIDS-compliant, including a "dataset_description.json" file. It assumes the dataset contains a derivatives
        folder with BOLD data preprocessed using a standard pipeline, specifically fMRIPrep. The pipeline directory
        must also include a "dataset_description.json" file for proper querying.

        The timeseries data of all subjects are appended to a single dictionary ``self.subject_timeseries``. Additional
        information regarding the structure of this dictionary can be found in the "Note" section.

        Basic BIDS directory:

        ::

            bids_root/
            ├── dataset_description.json
            ├── sub-<subject_label>/
            │   └── func/
            │       └── *task-*_events.tsv
            ├── derivatives/
            │   └── fmriprep-<version_label>/
            │       ├── dataset_description.json
            │       └── sub-<subject_label>/
            │           └── func/
            │               ├── *confounds_timeseries.tsv
            │               ├── *brain_mask.nii.gz
            │               └── *preproc_bold.nii.gz

        BIDS directory with session-level organization:

        ::

            bids_root/
            ├── dataset_description.json
            ├── sub-<subject_label>/
            │   └── ses-<session_label>/
            │       └── func/
            │           └── *task-*_events.tsv
            ├── derivatives/
            │   └── fmriprep-<version_label>/
            │       ├── dataset_description.json
            │       └── sub-<subject_label>/
            │           └── ses-<session_label>/
            │               └── func/
            │                   ├── *confounds_timeseries.tsv
            │                   ├── *brain_mask.nii.gz
            │                   └── *preproc_bold.nii.gz

        *Note: Only the preprocessed BOLD file is required. Additional files such as the confounds tsv (needed for
        denoising), mask, and task timing tsv file (needed for filtering a specific task condition) depend on the
        specific analyses. As mentioned previously, the "dataset_description.json" is required in both the bids root
        and pipeline directories for querying with pybids.*

        **This pipeline is most optimized for BOLD data preprocessed by fMRIPrep.**

        Parameters
        ----------
        bids_dir: :obj:`os.PathLike`
            Path to a BIDS compliant directory. A "dataset_description.json" file must be located in this directory
            or an error will be raised.

        task: :obj:`str`
            Name of task to extract timeseries data from (i.e "rest", "n-back", etc).

        session: :obj:`int`, :obj:`str`, or :obj:`None`, default=None
            Session ID to extract timeseries data from. Only a single session can be extracted at a time. While files
            having session IDs are not mandatory, this parameter must be specified if the dataset has multiple sessions
            . If ``session`` is None and multiple sessions are detected when the preprocessed NifTI files are queried,
            an error will be raised. The value can be an integer (e.g. ``session=2``) or a string (e.g.
            ``session="001"``).

        runs: :obj:`int`, :obj:`str`, :obj:`list[int]`, :obj:`list[str]`, or :obj:`None`, default=None
            List of run numbers to extract timeseries data from. Extracts all runs if unspecified. For instance,
            extract only "run-0" and "run-1", use ``runs=[0, 1]``. For non-integer run IDs, use strings:
            ``runs=["000", "001"]``.

        condition: :obj:`str` or :obj:`None`, default=None
            Isolates the timeseries data corresponding to a specific condition, only after the timeseries has been
            extracted and subjected to nuisance regression. Only a single condition can be extracted at a time.

        condition_tr_shift: :obj:`int`, default=0
            Number of TR units to units to offset both the start and end scan indices of a condition. This parameter
            only applies when a ``condition`` is specified. For more details about how this offset affects the
            calculation of task conditions, see the "Extraction of Task Conditions" section below.

            .. versionadded:: 0.20.0

        tr: :obj:`int`, :obj:`float` or :obj:`None`, default=None
            Repetition time (TR) for the specified task. If not provided, the TR will be automatically extracted from
            the first BOLD metadata file found for the task, searching first in the pipeline directory, then in
            the ``bids_dir`` if not found.

        slice_time_ref: :obj:`int` or :obj:`float`, default=0.0
            The reference slice expressed as a fraction of the ``tr`` that is subtracted from the condition onset times
            to adjust for slice time correction when ``condition`` is not None (``onset - slice_time_ref * tr``). Values
            can range from 0 to 1.

            .. versionadded:: 0.21.0

        run_subjects: :obj:`list[str]` or :obj:`None`, default=None
            List of subject IDs to process (e.g. ``run_subjects=["01", "02"]``). Processes all subjects if None.

        exclude_subjects: :obj:`list[str]` or :obj:`None`, default=None
            List of subject IDs to exclude (e.g. ``exclude_subjects=["01", "02"]``).

        exclude_niftis: :obj:`list[str]` or :obj:`None`, default=None
            List of the specific preprocessed NIfTI files to exclude, preventing their timeseries data from being
            extracted. Used if there are specific runs across different participants that need to be excluded.

            .. versionchanged:: 0.18.0 moved from being the second to last parameter, to being underneath
               ``exclude_subjects``

        pipeline_name: :obj:`str` or :obj:`None`, default=None
            The name of the pipeline folder in the derivatives folder containing the preprocessed data. If None,
            ``BIDSLayout`` will default to using the ``bids_dir`` with ``derivatives=True``. This parameter should be
            used if multiple pipelines exist or when the pipeline folder containing the "dataset_description.json" file
            is nested within another folder. The specified folder must contain the "dataset_description.json" file
            in its root level. For instance, if the json file is in "path/to/bids/derivatives/fmriprep/fmriprep-20.0.0",
            then ``pipeline_name = "fmriprep/fmriprep-20.0.0"``.

        n_cores: :obj:`int` or :obj:`None`, default=None
            The number of cores to use for multiprocessing with Joblib. The "loky" backend is used.

        parallel_log_config: :obj:`dict[str, Union[multiprocessing.Manager.Queue, int]]`
            Passes a user-defined managed queue and logging level to the internal timeseries extraction function
            when parallel processing (``n_cores``) is used. Note, if parallel processing is used, global logging
            configurations won't be passed to the child processes. Thus, to prevent the child processes from using the
            default logging behavior, this parameter must be used. Additionally, this parameter must be a dictionary
            and the available keys are:

            - "queue": The instance of ``multiprocessing.Manager.Queue`` to pass to ``QueueHandler``. If not specified,
              all logs will output to ``sys.stdout``.
            - "level": The logging level (e.g. ``logging.INFO``, ``logging.WARNING``). If not specified, the default
              level is ``logging.INFO``.

            Refer to the `neurocaps Logging Documentation <https://neurocaps.readthedocs.io/en/stable/logging.html>`_
            for a detailed example of setting up this parameter.

            .. versionchanged:: 0.18.0 moved from being the last parameter, to being underneath ``n_cores``

        verbose: :obj:`bool`, default=True
            If True, logs detailed subject-specific information including: subjects skipped due to missing required
            files, current subject being processed for timeseries extraction, confounds identified for nuisance
            regression in addition to requested confounds that are missing for a subject, and additional warnings
            encountered during the timeseries extraction process.

        flush: :obj:`bool`, default=False
            If True, flushes the logged subject-specific information produced during the timeseries extraction process.

        Returns
        -------
        self


            .. versionadded:: 0.19.3

        Note
        ----
        **Subject Timeseries Dictionary**: This method stores the extracted timeseries of all subjects
        in ``self.subject_timeseries``. The structure is a dictionary mapping subject IDs to their run IDs and
        their associated timeseries (TRs x ROIs) as a NumPy array:

        ::

            subject_timeseries = {
                    "101": {
                        "run-0": np.array([timeseries]), # Shape: TRs x ROIs
                        "run-1": np.array([timeseries]), # Shape: TRs x ROIs
                        "run-2": np.array([timeseries]), # Shape: TRs x ROIs
                    },
                    "102": {
                        "run-0": np.array([timeseries]), # Shape: TRs x ROIs
                        "run-1": np.array([timeseries]), # Shape: TRs x ROIs
                    }
                }

        By default, "run-0", will be used if run IDs are not specified in the NifTI file.

        **Parcellation & Nuisance Regression**: For timeseries extraction, nuisance regression, and spatial
        dimensionality reduction using a parcellation, Nilearn's ``NiftiLabelsMasker`` function is used. If requested,
        dummy scans are removed from the NIfTI images and confound dataset prior to timeseries extraction. For volumes
        exceeding a specified framewise displacement (FD) threshold, if the "use_sample_mask" key in the
        ``fd_threshold`` dictionary is set to True, then a boolean sample mask is generated (where False indicates the
        high motion volumes) and passed to the ``sample_mask`` parameter in Nilearn's ``NiftiLabelsMasker``. If,
        "use_sample_mask" key is False or not specified in the ``fd_threshold`` dictionary, then censoring is done
        after nuisance regression, which is the default behavior.

        **Extraction of Task Conditions**: The formula used for computing the scan indices corresponding to the
        corresponding to a specific condition:

        ::

            adjusted_onset = onset - slice_time_ref * tr
            adjusted_onset = adjusted_onset if adjusted_onset >= 0 else 0
            start_scan = int(adjusted_onset / tr) + condition_tr_shift
            end_scan = math.ceil((adjusted_onset + duration) / tr) + condition_tr_shift

        When partial scans are computed, ``int`` is used to round down for the beginning scan index and ``math.ceil``
        is used to round up for the ending scan index. Negative scan indices are set to 0 to avoid unintentional
        negative indexing. For simplicity, note that when ``slice_time_ref`` and ``condition_tr_shift`` are 0, the
        formula simplifies to:

        ::

            start_scan = int(onset / tr)
            end_scan = math.ceil((onset + duration) / tr)

        Filtering a specific condition from the timeseries is done after nuisance regression and the indices are used
        to extract the TRs corresponding to the condition from the timeseries. Additionally, if the "use_sample_mask"
        key in the ``fd_threshold`` dictionary is set to True, then the truncated 2D timeseries is temporarily padded
        to ensure the correct rows corresponding to the condition are obtained.
        """
        if runs and not isinstance(runs, list):
            runs = [runs]

        # Update attributes
        self._task_info = {
            "task": task,
            "session": session,
            "runs": runs,
            "condition": condition,
            "condition_tr_shift": condition_tr_shift,
            "tr": tr,
            "slice_time_ref": slice_time_ref,
        }

        # Quick check condition_tr_shift and slice_time_ref to not cause issues later in the pipeline
        self._validate_get_bold_params()

        # Initialize new attributes
        self._subject_ids = []
        self._subject_timeseries = {}
        self._subject_info = {}
        self._n_cores = n_cores

        layout = self._call_layout(bids_dir, pipeline_name)
        subj_ids = sorted(layout.get(return_type="id", target="subject", task=task, space=self._space, suffix="bold"))

        if not subj_ids:
            msg = (
                "No subject IDs found - potential reasons: "
                "1. Incorrect template space (default: 'MNI152NLin2009cAsym'). "
                "Fix: Set correct template space using `self.space = 'TEMPLATE_SPACE'` (e.g. 'MNI152NLin6Asym') "
                "2. Incorrect task name specified in `task` parameter."
            )
            raise BIDSQueryError(msg)

        if exclude_subjects:
            subj_ids = sorted([subj_id for subj_id in subj_ids if subj_id not in map(str, exclude_subjects)])

        if run_subjects:
            subj_ids = sorted([subj_id for subj_id in subj_ids if subj_id in map(str, run_subjects)])

        # Setup extraction
        self._setup_extraction(layout, subj_ids, exclude_niftis, verbose)

        if self._n_cores:
            # Generate list of tuples for each subject
            args_list = [
                (
                    subj_id,
                    self._subject_info[subj_id]["prepped_files"],
                    self._subject_info[subj_id]["run_list"],
                    self._parcel_approach,
                    self._signal_clean_info,
                    self._task_info,
                    self._subject_info[subj_id]["tr"],
                    verbose,
                    flush,
                    parallel_log_config,
                )
                for subj_id in self._subject_ids
            ]

            parallel = Parallel(return_as="generator", n_jobs=self._n_cores, backend="loky")
            outputs = parallel(delayed(_extract_timeseries)(*args) for args in args_list)

            for output in outputs:
                if isinstance(output, dict):
                    self._subject_timeseries.update(output)
        else:
            if parallel_log_config:
                LG.warning(
                    "`parallel_log_config` is only used for parallel processing. The default logger can be "
                    "modified by configuring either the root logger or a logger for a specific module prior to "
                    "package import."
                )

            for subj_id in self._subject_ids:
                subject_timeseries = _extract_timeseries(
                    subj_id=subj_id,
                    **self._subject_info[subj_id],
                    parcel_approach=self._parcel_approach,
                    signal_clean_info=self._signal_clean_info,
                    task_info=self._task_info,
                    verbose=verbose,
                    flush=flush,
                )

                # Aggregate new timeseries
                if isinstance(subject_timeseries, dict):
                    self._subject_timeseries.update(subject_timeseries)

        return self

    def _validate_get_bold_params(self):
        if self.task_info["condition_tr_shift"] != 0:
            if not isinstance(self.task_info["condition_tr_shift"], int) or self.task_info["condition_tr_shift"] < 0:
                raise ValueError("`condition_tr_shift` must be a integer value equal to or greater than 0.")

            if self.task_info["condition"] is None:
                LG.warning("`condition_tr_shift` specified but `condition` is None.")

        if self.task_info["slice_time_ref"] != 0:
            if not isinstance(self.task_info["slice_time_ref"], (float, int)) or (
                self.task_info["slice_time_ref"] < 0 or (self.task_info["slice_time_ref"] > 1)
            ):
                raise ValueError("`slice_time_ref` must be a numerical value from 0 to 1.")

            if self.task_info["condition"] is None:
                LG.warning("`slice_time_ref` specified but `condition` is None.")

    @staticmethod
    @lru_cache(maxsize=4)
    def _call_layout(bids_dir, pipeline_name):
        try:
            import bids
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "This function relies on the pybids package to query subject-specific files. "
                "If on Windows, pybids does not install by default to avoid long path error issues "
                "during installation. Try using `pip install pybids` or `pip install neurocaps[windows]`."
            )

        bids_dir = os.path.normpath(bids_dir).rstrip(os.path.sep)

        if bids_dir.endswith("derivatives"):
            bids_dir = os.path.dirname(bids_dir)

        if pipeline_name:
            pipeline_name = os.path.normpath(pipeline_name).lstrip(os.path.sep).rstrip(os.path.sep)

            if pipeline_name.startswith("derivatives"):
                pipeline_name = pipeline_name[len("derivatives") :].lstrip(os.path.sep)

            layout = bids.BIDSLayout(bids_dir, derivatives=os.path.join(bids_dir, "derivatives", pipeline_name))
        else:
            layout = bids.BIDSLayout(bids_dir, derivatives=True)

        LG.info(f"{layout}")

        return layout

    # Get valid subjects to iterate through
    def _setup_extraction(self, layout, subj_ids, exclude_niftis, verbose):
        base_dict = {"layout": layout, "subj_id": None}

        for subj_id in subj_ids:
            base_dict["subj_id"] = subj_id
            files = self._build_dict(base_dict)

            # Remove excluded file from the niftis list, which will prevent it from being processed
            if exclude_niftis and files["niftis"]:
                files["niftis"] = self._exclude(files["niftis"], exclude_niftis)

            # Get subject header
            subject_header = self._header(subj_id)

            # Check files
            skip, msg = self._check_files(files)

            if msg and verbose:
                LG.warning(subject_header + msg)
            if skip:
                continue

            # Ensure only a single session is present if session is None
            if not self._task_info["session"]:
                self._check_sess(files["niftis"], subject_header)

            # Generate a list of runs to iterate through based on runs in niftis
            check_runs = self._gen_runs(files["niftis"])

            if check_runs:
                run_list = self._intersect_runs(check_runs, files)

                # Skip subject if no run has all needed files present
                if not run_list:
                    if verbose:
                        LG.warning(
                            f"{subject_header}"
                            "Timeseries Extraction Skipped: None of the necessary files (i.e NifTIs, masks, "
                            "confound tsv files, confound json files, event files) are from the same run."
                        )
                    continue

                if len(run_list) != len(check_runs):
                    if verbose:
                        LG.warning(
                            f"{subject_header}"
                            "Only the following runs available contain all required files: "
                            f"{', '.join(run_list)}."
                        )
            else:
                # Allows for nifti files that do not have the run- description
                run_list = [None]

            # Add subject list to subject attribute. These are subjects that will be ran
            self._subject_ids.append(subj_id)

            # Get repetition time for the subject
            tr = self._get_tr(files["bold_meta"], subject_header, verbose)

            # Store subject specific information
            self._subject_info[subj_id] = {"prepped_files": files, "tr": tr, "run_list": run_list}

    def _get_files(
        self, layout, extension, subj_id, scope="derivatives", suffix=None, desc=None, event=False, space="attr"
    ):
        query_dict = {
            "scope": scope,
            "return_type": "file",
            "task": self._task_info["task"],
            "extension": extension,
            "subject": subj_id,
        }

        if desc:
            query_dict.update({"desc": desc})
        if suffix:
            query_dict.update({"suffix": suffix})
        if self._task_info["session"]:
            query_dict.update({"session": self._task_info["session"]})
        if not event and not desc:
            query_dict.update({"space": self._space if space == "attr" else space})

        return sorted(layout.get(**query_dict))

    def _build_dict(self, base):
        files = {}
        files["niftis"] = self._get_files(**base, suffix="bold", extension="nii.gz")
        files["masks"] = self._get_files(**base, suffix="mask", extension="nii.gz")
        files["bold_meta"] = self._get_files(**base, suffix="bold", extension="json")

        if not files["bold_meta"]:
            files["bold_meta"] = self._get_files(**base, scope="raw", suffix="bold", extension="json", space=None)
        if self._task_info["condition"]:
            files["events"] = self._get_files(**base, scope="raw", suffix="events", extension="tsv", event=True)
        else:
            files["events"] = []

        files["confounds"] = self._get_files(**base, desc="confounds", extension="tsv")
        files["confounds_metas"] = self._get_files(**base, extension="json", desc="confounds")

        return files

    @staticmethod
    def _exclude(niftis, exclude_niftis):
        return [nifti for nifti in niftis if os.path.basename(nifti) not in exclude_niftis]

    def _header(self, subj_id):
        # Base subject message
        sub_message = f'[SUBJECT: {subj_id} | SESSION: {self._task_info["session"]} | TASK: {self._task_info["task"]}]'
        subject_header = f"{sub_message} "
        return subject_header

    def _check_files(self, files):
        skip, msg = None, None

        if not files["niftis"]:
            skip, msg = (
                True,
                "Timeseries Extraction Skipped: No NifTI files were found or all NifTI files were excluded.",
            )

        if not files["masks"]:
            skip, msg = False, "Missing mask file but timeseries extraction will continue."

        if self._signal_clean_info["use_confounds"]:
            if not files["confounds"]:
                skip, msg = (
                    True,
                    "Timeseries Extraction Skipped: `use_confounds` is requested but no confound files found.",
                )

            if not files["confounds_metas"] and self._signal_clean_info["n_acompcor_separate"]:
                skip = True
                msg = (
                    "Timeseries Extraction Skipped: No confound metadata file found, which is needed to locate the "
                    "first n components of the white-matter and cerebrospinal fluid masks separately."
                )

            if self._task_info["condition"] and not files["events"]:
                skip, msg = True, "Timeseries Extraction Skipped: `condition` is specified but no event files found."

        return skip, msg

    def _check_sess(self, niftis, subject_header):
        ses_list = []

        for nifti in niftis:
            if "ses-" in os.path.basename(nifti):
                ses_list.append(re.search(r"ses-(\S+?)[-_]", os.path.basename(nifti))[0][:-1])

        ses_list = set(ses_list)

        if len(ses_list) > 1:
            raise ValueError(
                f"{subject_header}"
                "`session` not specified but subject has more than one session: "
                f"{', '.join(ses_list)}\n" + "In order to continue timeseries extraction, the "
                "specific session to extract must be specified using `session`."
            )

    def _gen_runs(self, niftis):
        check_runs = []

        for nifti in niftis:
            if self._task_info["runs"]:
                check_runs.extend([f"run-{run}" for run in self._task_info["runs"]])
            else:
                if "run-" in os.path.basename(niftis[0]):
                    check_runs.append(re.search(r"run-(\S+?)[-_]", os.path.basename(nifti))[0][:-1])

        return check_runs

    def _intersect_runs(self, check_runs, files):
        run_list = []

        # Check if at least one run has all files present
        for run in check_runs:
            curr_list = []

            # Assess is any of these returns True
            curr_list.append(any([run in file for file in files["niftis"]]))
            if self._task_info["condition"]:
                curr_list.append(any([run in file for file in files["events"]]))
            if self._signal_clean_info["use_confounds"]:
                curr_list.append(any([run in file for file in files["confounds"]]))
                if self._signal_clean_info["n_acompcor_separate"]:
                    curr_list.append(any([run in file for file in files["confounds_metas"]]))

            if files["masks"]:
                curr_list.append(any([run in file for file in files["masks"]]))
            # Append runs that contain all needed files
            if all(curr_list):
                run_list.append(run)

        return run_list

    def _get_tr(self, bold_meta, subject_header, verbose):
        try:
            if self._task_info["tr"]:
                tr = self._task_info["tr"]
            else:
                with open(bold_meta[0], "r") as json_file:
                    tr = json.load(json_file)["RepetitionTime"]
        except (IndexError, KeyError, json.JSONDecodeError) as e:
            base_msg = "`tr` not specified and could not be extracted since using the first BOLD metadata file"
            base_msg += " due to " + (
                f"there being no BOLD metadata files for [TASK: {self._task_info['task']}]"
                if str(type(e).__name__) == "IndexError"
                else f"{type(e).__name__} - {str(e)}"
            )

            if self._task_info["condition"]:
                raise ValueError(
                    f"{subject_header}" f"{base_msg}" + " The `tr` must be given when `condition` is specified."
                )
            elif any(
                [
                    self._signal_clean_info["masker_init"]["high_pass"],
                    self._signal_clean_info["masker_init"]["low_pass"],
                ]
            ):
                raise ValueError(
                    f"{subject_header}"
                    f"{base_msg}" + " The `tr` must be given when `high_pass` or `low_pass` is specified."
                )
            else:
                tr = None

                if verbose:
                    LG.warning(
                        f"{subject_header}" f"{base_msg}" + " `tr` has been set to None but extraction will continue."
                    )

        return tr

    @staticmethod
    def _raise_error(msg):
        raise AttributeError(
            f"{msg} since `self.subject_timeseries` is None, either run "
            "`self.get_bold()` or assign a valid timeseries dictionary to `self.subject_timeseries`."
        )

    def timeseries_to_pickle(self, output_dir: Union[str, os.PathLike], filename: Optional[str] = None) -> None:
        """
        Save the Extracted Subject Timeseries.

        Saves the extracted timeseries stored in the ``self.subject_timeseries`` dictionary (obtained from running
        ``self.get_bold``) as a pickle file. This allows for data persistence and easy conversion back into
        dictionary form for later use.

        Parameters
        ----------
        output_dir: :obj:`os.PathLike`
            Directory to save ``self.subject_timeseries`` dictionary as a pickle file. The directory will be created if
            it does not exist.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file with or without the "pkl" extension.

            .. versionchanged:: 0.19.0  ``file_name`` to ``filename``

        Returns
        -------
        self


            .. versionadded:: 0.19.3
        """
        if not self.subject_timeseries:
            self._raise_error("Cannot save pickle file")

        if output_dir and not os.path.exists(output_dir):
            os.makedirs(output_dir)

        if filename is None:
            save_filename = "subject_timeseries.pkl"
        else:
            save_filename = f"{os.path.splitext(filename.rstrip())[0].rstrip()}.pkl"

        with open(os.path.join(output_dir, save_filename), "wb") as f:
            dump(self._subject_timeseries, f)

        return self

    def visualize_bold(
        self,
        subj_id: Union[int, str],
        run: Union[int, str],
        roi_indx: Optional[Union[Union[int, str], Union[list[str], list[int]]]] = None,
        region: Optional[str] = None,
        show_figs: bool = True,
        output_dir: Optional[Union[str, os.PathLike]] = None,
        filename: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Plot the Extracted Subject Timeseries.

        Uses the ``self.subject_timeseries`` to visualize the extracted BOLD timeseries data of  data Regions of
        Interest (ROIs) or regions for a specific subject and run.

        Parameters
        ----------
        subj_id: :obj:`str` or :obj:`int`
            The ID of the subject.

        run: :obj:`int` or :obj:`str`
            The run ID of the subject to plot.

        roi_indx: :obj:`int`, :obj:`str`, :obj:`list[int]`, :obj:`list[int]` or :obj:`None`, default=None
            The indices of the parcellation nodes to plot. See "nodes" in ``self.parcel_approach`` for valid
            nodes.

        region: :obj:`str` or :obj:`None`, default=None
            The region of the parcellation to plot. If not None, all nodes in the specified region will be averaged
            then plotted. See "regions" in ``self.parcel_approach`` for valid region.

        show_figs: :obj:`bool`, default=True
            Display figures.

        output_dir: :obj:`os.PathLike` or :obj:`None`, default=None
            Directory to save plot as png image. The directory will be created if it does not exist. If None, plot will
            not be saved.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file without the extension.

            .. versionchanged:: 0.19.0  ``file_name`` to ``filename``

        kwargs: :obj:`dict`
            Keyword arguments used when saving figures. Valid keywords include:

            - dpi: :obj:`int`, default=300
                Dots per inch for the figure. Default is 300 if ``output_dir`` is provided and ``dpi`` is not
                specified.
            - figsize: :obj:`tuple`, default=(11, 5)
                Size of the figure in inches. Default is (11, 5) if ``figsize`` is not specified.
            - bbox_inches: :obj:`str` or :obj:`None`, default="tight"
                Alters size of the whitespace in the saved image.

        Returns
        -------
        self


            .. versionadded:: 0.19.3

        Note
        ----
        **Parcellation Approach**: the "nodes" and "regions" sub-keys are required in ``parcel_approach``.
        """

        if not self.subject_timeseries:
            self._raise_error("Cannot plot bold data")

        if roi_indx is None and region is None:
            raise ValueError("either `roi_indx` or `region` must be specified.")

        if roi_indx is not None and region is not None:
            raise ValueError("`roi_indx` and `region` can not be used simultaneously.")

        if filename is not None and output_dir is None:
            LG.warning("`filename` supplied but no `output_dir` specified. Files will not be saved.")

        parcellation_name = list(self._parcel_approach)[0]

        # Defaults
        defaults = {"dpi": 300, "figsize": (11, 5), "bbox_inches": "tight"}

        plot_dict = _check_kwargs(defaults, **kwargs)

        # Obtain the column indices associated with the rois
        if roi_indx or roi_indx == 0:
            plot_indxs = self._get_roi_indices(roi_indx, parcellation_name)
        else:
            plot_indxs = self._get_region_indices(region, parcellation_name)

        plt.figure(figsize=plot_dict["figsize"])

        timeseries = self._subject_timeseries[str(subj_id)][f"run-{run}"]

        if roi_indx or roi_indx == 0:
            plt.plot(range(1, timeseries.shape[0] + 1), timeseries[:, plot_indxs])

            if isinstance(roi_indx, (int, str)) or (isinstance(roi_indx, list) and len(roi_indx) == 1):
                if isinstance(roi_indx, int):
                    roi_title = self._parcel_approach[parcellation_name]["nodes"][roi_indx]
                elif isinstance(roi_indx, str):
                    roi_title = roi_indx
                else:
                    roi_title = roi_indx[0]
                plt.title(roi_title)
        else:
            plt.plot(range(1, timeseries.shape[0] + 1), np.mean(timeseries[:, plot_indxs], axis=1))
            plt.title(region)

        plt.xlabel("TR")

        if output_dir:
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
            if filename:
                save_filename = f"{os.path.splitext(filename.rstrip())[0].rstrip()}.png"
            else:
                save_filename = f"subject-{subj_id}_run-{run}_timeseries.png"

            plt.savefig(
                os.path.join(output_dir, save_filename), dpi=plot_dict["dpi"], bbox_inches=plot_dict["bbox_inches"]
            )

        plt.show() if show_figs else plt.close()

        return self

    def _get_roi_indices(self, roi_indx, parcellation_name):
        if isinstance(roi_indx, int):
            plot_indxs = roi_indx
        elif isinstance(roi_indx, str):
            # Check if parcellation_approach is custom
            if "Custom" in self._parcel_approach and "nodes" not in self._parcel_approach["Custom"]:
                _check_parcel_approach(parcel_approach=self._parcel_approach, call="visualize_bold")

            plot_indxs = list(self._parcel_approach[parcellation_name]["nodes"]).index(roi_indx)
        else:
            if all([isinstance(indx, int) for indx in roi_indx]):
                plot_indxs = np.array(roi_indx)
            elif all([isinstance(indx, str) for indx in roi_indx]):
                # Check if parcellation_approach is custom
                if "Custom" in self._parcel_approach and "nodes" not in self._parcel_approach["Custom"]:
                    _check_parcel_approach(parcel_approach=self._parcel_approach, call="visualize_bold")

                plot_indxs = np.array(
                    [list(self._parcel_approach[parcellation_name]["nodes"]).index(index) for index in roi_indx]
                )
            else:
                raise ValueError("All elements in `roi_indx` need to be all strings or all integers.")

        return plot_indxs

    def _get_region_indices(self, region, parcellation_name):
        if "Custom" in self._parcel_approach:
            if "regions" not in self._parcel_approach["Custom"]:
                _check_parcel_approach(parcel_approach=self._parcel_approach, call="visualize_bold")
            else:
                plot_indxs = np.array(
                    list(self._parcel_approach["Custom"]["regions"][region]["lh"])
                    + list(self._parcel_approach["Custom"]["regions"][region]["rh"])
                )
        else:
            plot_indxs = np.array(
                [
                    index
                    for index, label in enumerate(self._parcel_approach[parcellation_name]["nodes"])
                    if region in label
                ]
            )

        return plot_indxs



import os
from typing import Optional, Union

import matplotlib.pyplot as plt, pandas as pd

from .._utils import _check_kwargs, _create_display, _logger, _save_contents

LG = _logger(__name__)


def transition_matrix(
    trans_dict: dict[str, pd.DataFrame],
    output_dir: Optional[os.PathLike] = None,
    suffix_title: Optional[str] = None,
    suffix_filename: Optional[str] = None,
    show_figs: bool = True,
    save_plots: bool = True,
    return_df: bool = True,
    save_df: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """
    Generate and Visualize the Averaged Transition Probabilities.

    Uses the "transition_probability" output from ``CAP.calculate_metrics`` to generate and visualize the averaged
    transition probability matrix for all groups from the analysis.

    Parameters
    ----------
    trans_dict: :obj: `dict[str, pd.DataFrame]`
        A dictionary mapping groups to pandas DataFrame containing the transition probabilities for each subject.
        This assumes the output from ``CAP.calculate_metrics`` is being used, specifically
        ``metrics_output["transition_probability"]``.

    output_dir: :obj:`os.PathLike` or :obj:`None`, default=None
        Directory to save plots (if ``save_plots`` is True) and transition probability matrices DataFrames (if
        ``save_df`` is True) to. The directory will be created if it does not exist. Plots and dataframes will not
        be saved if None.

    suffix_title: :obj:`str` or :obj:`None`, default=None
        Appended to the title of each plot.

    suffix_filename: :obj:`str` or :obj:`None`, default=None
        Appended to the filename of each saved plot if ``output_dir`` is provided.

        .. versionadded:: 0.19.0

    show_figs: :obj:`bool`, default=True
        Display figures.

    save_plots: :obj:`bool`, default=True
        If True, plots are saves as png images. For this to be used, ``output_dir`` must be specified.

    return_df: :obj:`bool`, default=False
        If True, returns a dictionary with a transition probability matrix for each group.

    save_df: :obj:`bool`, default=False,
        If True, saves the transition probability matrix contained in the DataFrames as csv files. For this to be used,
        ``output_dir`` must be specified.

    kwargs: :obj:`dict`
        Keyword arguments used when modifying figures. Valid keywords include:

        - dpi: :obj:`int`, default=300
            Dots per inch for the figure. Default is 300 if ``output_dir`` is provided and ``dpi`` is not
            specified.
        - figsize: :obj:`tuple`, default=(8, 6)
            Size of the figure in inches.
        - fontsize: :obj:`int`, default=14
            Font size for the plot title, x-axis title, and y-axis title of each plot.
        - xticklabels_size: :obj:`int`, default=8
            Font size for x-axis tick labels.
        - yticklabels_size: :obj:`int`, default=8
            Font size for y-axis tick labels.
        - shrink: :obj:`float`, default=0.8
            Fraction by which to shrink the colorbar.
        - cbarlabels_size: :obj:`int`, default=8
            Font size for the colorbar labels.
        - xlabel_rotation: :obj:`int`, default=0
            Rotation angle for x-axis labels.
        - ylabel_rotation: :obj:`int`, default=0
            Rotation angle for y-axis labels.
        - annot: :obj:`bool`, default=False
            Add values to each cell.
        - annot_kws: :obj:`dict`, default=None,
            Customize the annotations.
        - fmt: :obj:`str`, default=".2g",
            Modify how the annotated vales are presented.
        - linewidths: :obj:`float`, default=0
            Padding between each cell in the plot.
        - borderwidths: :obj:`float`, default=0
            Width of the border around the plot.
        - linecolor: :obj:`str`, default="black"
            Color of the line that separates each cell.
        - edgecolors: :obj:`str` or :obj:`None`, default=None
            Color of the edges.
        - alpha: :obj:`float` or :obj:`None`, default=None
            Controls transparency and ranges from 0 (transparent) to 1 (opaque).
        - bbox_inches: :obj:`str` or :obj:`None`, default="tight"
            Alters size of the whitespace in the saved image.
        - cmap: :obj:`str`, :obj:`callable` default="coolwarm"
            Color map for the plot cells. For this parameter, you can use pre-made color palettes or create custom ones.
            Below is a list of valid options:

            - Strings to call seaborn's pre-made palettes.
            - ``seaborn.diverging_palette`` function to generate custom palettes.
            - ``matplotlib.color.LinearSegmentedColormap`` to generate custom palettes.

        - vmin: :obj:`float` or :obj:`None`, default=None
                The minimum value to display in colormap.
        - vmax: :obj:`float` or :obj:`None`, default=None
                The maximum value to display in colormap.

    Returns
    -------
    dict[str, pd.DataFrame]
        An instance of a pandas DataFrame for each group.

    Note
    ----
    **Dataframe Representation**: Rows represent "from" and columns represent "to". For instance,
    the probability at ``df.loc["CAP-1", "CAP-2"]`` represents the averaged probability from transitioning from
    CAP-1 to CAP-2.

    +------------+---------+-------+-------+
    | From/To    |  CAP-1  | CAP-2 | CAP-3 |
    +============+=========+=======+=======+
    | CAP-1      |  0.40   | 0.35  | 0.25  |
    +------------+---------+-------+-------+
    | CAP-2      |  0.20   | 0.45  | 0.35  |
    +------------+---------+-------+-------+
    | CAP-3      |  0.35   | 0.18  |  0.47 |
    +------------+---------+-------+-------+

    *The "From/To" index name appears in versions >=0.17.11.*
    """
    assert isinstance(trans_dict, dict), "transition_dict must be in the form dict[str, pd.DataFrame]."

    if suffix_filename is not None and output_dir is None:
        LG.warning("`suffix_filename` supplied but no `output_dir` specified. Files will not be saved.")

    # Create plot dictionary
    defaults = {
        "dpi": 300,
        "figsize": (8, 6),
        "fontsize": 14,
        "xticklabels_size": 8,
        "yticklabels_size": 8,
        "shrink": 0.8,
        "cbarlabels_size": 8,
        "xlabel_rotation": 0,
        "ylabel_rotation": 0,
        "annot": False,
        "linewidths": 0,
        "linecolor": "black",
        "cmap": "coolwarm",
        "fmt": ".2g",
        "borderwidths": 0,
        "edgecolors": None,
        "alpha": None,
        "bbox_inches": "tight",
        "annot_kws": None,
        "vmin": None,
        "vmax": None,
    }

    plot_dict = _check_kwargs(defaults, **kwargs)

    trans_mat_dict = {}

    for group in trans_dict:
        df = trans_dict[group]
        # Get indices and averaged probabilities
        indices, averaged_probabilities = df.iloc[:, 3:].mean().index, df.iloc[:, 3:].mean().values
        # Get the maximum CAP
        max_cap = str(max([float(i) for i in indices])).split(".")[0]
        cap_names = [f"CAP-{num}" for num in range(1, int(max_cap) + 1)]
        trans_mat = pd.DataFrame(index=cap_names, columns=cap_names, dtype="float64")
        # Add name to index
        trans_mat.index.name = "From/To"

        # Create matrix
        for location, name in enumerate(indices):
            trans_mat.loc[f"CAP-{name.split('.')[0]}", f"CAP-{name.split('.')[1]}"] = averaged_probabilities[location]

        display = _create_display(trans_mat, plot_dict, suffix_title, group, "trans")

        # Store df in dict
        trans_mat_dict[group] = trans_mat

        # Save figure & dataframe
        if output_dir:
            _save_contents(
                output_dir, suffix_filename, group, trans_mat_dict, plot_dict, save_plots, save_df, display, "trans"
            )

        # Display figures
        plt.show() if show_figs else plt.close()

    if return_df:
        return trans_mat_dict



import copy, os
from typing import Union, Optional

import numpy as np
from numpy.typing import NDArray

from .._utils import _convert_pickle_to_dict, _dicts_to_pickles, _logger

LG = _logger(__name__)


def merge_dicts(
    subject_timeseries_list: Union[list[dict[str, dict[str, NDArray[np.floating]]]], list[os.PathLike]],
    return_merged_dict: bool = True,
    return_reduced_dicts: bool = False,
    output_dir: Optional[Union[str, os.PathLike]] = None,
    filenames: Optional[list[str]] = None,
    save_reduced_dicts: bool = False,
) -> dict[str, dict[str, dict[str, NDArray[np.floating]]]]:
    """
    Merge Participant Timeseries Across Multiple Sessions or Tasks.

    Merge subject timeseries data across dictionaries, concatenating matching run IDs. Only subjects present across all
    input dictionaries are included in the merged output.

    For example, if three dictionaries are provided contain subject 1 with:

        - dict 1: run-1 (resting-state)
        - dict 2: run-1 and run-2 (stroop)
        - dict 3: run-3 (n-back)

    Then subject 1 in the final merged dictionary will contain:

        - run-1: concatenated timeseries from dict 1 and dict 2 (resting-state + stroop)
        - run-2: timeseries from dict 2 (stroop)
        - run-3: timeseries from dict 3 (n-back)

    This function supports workflows for identifying similar CAPs across tasks or sessions. Specifically, using the
    merged dictionary as input for ``CAP.get_caps`` and the filtered input dictionaries, containing only subjects
    present in the merged dictionary, as inputs for ``CAP.calculate_metrics`` to compute participant-wise CAP metrics
    for each task.

    Parameters
    ----------
    subject_timeseries_list: :obj:`list[dict[str, dict[str, np.ndarray]]]` or :obj:`list[os.PathLike]`
        A list where each element consist of a dictionary mapping subject IDs to their run IDs and associated
        timeseries (TRs x ROIs) as a NumPy array. Can also be a list consisting of paths to pickle files
        containing this same structure. The expected structure of each dictionary is as follows:

        ::

            subject_timeseries = {
                    "101": {
                        "run-0": np.array([...]), # Shape: TRs x ROIs
                        "run-1": np.array([...]), # Shape: TRs x ROIs
                        "run-2": np.array([...]), # Shape: TRs x ROIs
                    },
                    "102": {
                        "run-0": np.array([...]), # Shape: TRs x ROIs
                        "run-1": np.array([...]), # Shape: TRs x ROIs
                    }
                }

    return_merged_dict: :obj:`bool`, default=True
        If True, returns a single dictionary containing the merged dictionary under a key named "merged".

    return_reduced_dicts: :obj:`bool`, default=False
        If True, returns a single dictionary containing the input dictionaries filtered to only include subjects present
        in the merged dictionary. Keys are named "dict_{0}" where {0} corresponds to the dictionary's position in the
        input list.

    output_dir: :obj:`os.PathLike` or :obj:`None`, default=None
        Directory to save the merged or reduced dictionaries as pickle files. The directory will be created
        if it does not exist. For the reduced dictionaries to be saved, ``save_reduced_dicts`` must be set to True.
        If ``save_reduced_dicts`` is False and ``output_dir`` is provided, only the merged dictionary will be saved.
        Dictionaries will not be saved if None.

    filenames: :obj:`list[str]` or :obj:`None`, default=None
        A list of file names for saving dictionaries when ``output_dir`` is provided. If ``save_reduced_dicts`` is False,
        only a list with a single name should be supplied, which will be used to save the merged dictionary. If
        ``save_reduced_dicts`` is True, provide `N+1` names (where `N` is the length of subject_timeseries_list) - `N`
        names for individual reduced dictionaries followed by one name for the merged dictionary. Names are assigned by
        input position order. Full paths are handled using basename, and extensions are ignored. If None, uses default
        names - "subject_timeseries_{0}_reduced.pkl" (where {0} indicates the original input order) and
        "merged_subject_timeseries.pkl" for the merged dictionary.

        .. versionchanged:: 0.19.0  ``file_names`` to ``filenames``

    save_reduced_dicts: :obj:`bool` or None, default=False
        If True and the ``output_dir`` is provided, then the reduced dictionaries are saved as pickle files.

    Returns
    -------
    dict[str, dict[str, dict[str, np.ndarray]]]
        A nested dictionary containing the merged subject timeseries and reduced subject timeseries if specified.

    References
    ----------
    Kupis, L., Romero, C., Dirks, B., Hoang, S., Parladé, M. V., Beaumont, A. L., Cardona, S. M., Alessandri, M., Chang, C., Nomi, J. S., & Uddin, L. Q. (2020). Evoked and intrinsic brain network dynamics in children with autism spectrum disorder. NeuroImage: Clinical, 28, 102396. https://doi.org/10.1016/j.nicl.2020.102396
    """

    assert isinstance(subject_timeseries_list, list), "`subject_timeseries_list` must be a list."
    assert len(subject_timeseries_list) > 1, "Merging cannot be done with less than two dictionaries or files."

    if filenames is not None and output_dir is None:
        LG.warning("`filenames` supplied but no `output_dir` specified. Files will not be saved.")

    if isinstance(subject_timeseries_list[0], dict):
        subject_timeseries_merged = subject_timeseries_list[0]
    else:
        subject_timeseries_merged = _convert_pickle_to_dict(subject_timeseries_list[0])

    # Get common subject ids
    subject_set = {}

    for curr_dict in subject_timeseries_list:
        if isinstance(curr_dict, str) and curr_dict.endswith(".pkl"):
            curr_dict = _convert_pickle_to_dict(curr_dict)

        if not subject_set:
            subject_set = set(curr_dict)

        subject_set = subject_set.intersection(list(curr_dict))

    # Order subjects
    intersect_subjects = sorted(list(subject_set))

    subject_timeseries_merged = {}

    for curr_dict in subject_timeseries_list:
        if isinstance(curr_dict, str) and curr_dict.endswith(".pkl"):
            curr_dict = _convert_pickle_to_dict(curr_dict)

        for subj_id in intersect_subjects:
            if subj_id not in subject_timeseries_merged:
                subject_timeseries_merged.update({subj_id: {}})
            # Get run names in the current iteration
            for curr_run in curr_dict[subj_id]:
                # If run is in merged dict, stack. If not, add
                if curr_run in subject_timeseries_merged[subj_id]:
                    subject_timeseries_merged[subj_id][curr_run] = np.vstack(
                        [subject_timeseries_merged[subj_id][curr_run], curr_dict[subj_id][curr_run]]
                    )
                else:
                    subject_timeseries_merged[subj_id].update({curr_run: curr_dict[subj_id][curr_run]})

            # Sort runs lexicographically
            if list(subject_timeseries_merged[subj_id]) != sorted(subject_timeseries_merged[subj_id].keys()):
                subject_timeseries_merged[subj_id] = {
                    run_id: subject_timeseries_merged[subj_id][run_id]
                    for run_id in sorted(subject_timeseries_merged[subj_id].keys())
                }

    modified_dicts = {}

    if return_reduced_dicts or (save_reduced_dicts and output_dir):
        for indx, curr_dict in enumerate(subject_timeseries_list):
            if "pkl" in curr_dict:
                curr_dict = _convert_pickle_to_dict(curr_dict)
            else:
                curr_dict = copy.deepcopy(curr_dict)

            if any([elem in subject_timeseries_merged for elem in curr_dict]):
                modified_dicts[f"dict_{indx}"] = {}
                for subj_id in subject_timeseries_merged:
                    if subj_id in curr_dict:
                        modified_dicts[f"dict_{indx}"].update({subj_id: curr_dict[subj_id]})

    if return_merged_dict or output_dir:
        modified_dicts["merged"] = subject_timeseries_merged

    if output_dir:
        message = (
            "Length of `file_names` must be equal to 1 if `save_reduced_dicts`is False or the length of "
            "`subject_timeseries_list` + 1 if `save_reduced_dicts` is True."
        )

        _dicts_to_pickles(
            output_dir=output_dir,
            dict_list=modified_dicts,
            call="merge",
            filenames=filenames,
            message=message,
            save_reduced_dicts=save_reduced_dicts,
        )

    if return_merged_dict or return_reduced_dicts:
        return modified_dicts



"""Internal logging function and class for flushing"""

import logging, sys
from logging.handlers import QueueHandler

# Global variables to determine if a handler is user defined or defined by OS
_USER_ROOT_HANDLER = None
_USER_MODULE_HANDLERS = {}


class _Flush(logging.StreamHandler):
    def emit(self, record):
        super().emit(record)
        self.flush()


def _logger(name, flush=False, top_level=True, parallel_log_config=None):
    global _USER_ROOT_HANDLER, _USER_MODULE_HANDLERS

    logger = logging.getLogger(name)
    parallel_module = "neurocaps._utils.extraction.extract_timeseries"

    # Windows appears to assign stderr has the root handler after the top level loggers are assigned, which causes
    # any loggers not assigned at top level to adopt this handler. Global variable used to assess if the base root
    # handler is user defined or assigned by the system
    if top_level is True:
        _USER_ROOT_HANDLER = logging.getLogger().hasHandlers()
        _USER_MODULE_HANDLERS[logger.name] = logger.handlers

    # Special case for parallel config to pass a user-defined logger specifically for parallel processing.
    if logger.name == parallel_module and not top_level and parallel_log_config:
        # Only QueueHandler will be in handler list
        logger.handlers.clear()
        if "queue" in parallel_log_config:
            logger.addHandler(QueueHandler(parallel_log_config["queue"]))
        if "level" in parallel_log_config:
            logger.setLevel(parallel_log_config["level"])

    if not logger.level:
        logger.setLevel(logging.INFO)

    # Check if user defined root handler or assigned a specific handler for module
    default_handlers = _USER_ROOT_HANDLER or _USER_MODULE_HANDLERS[logger.name]

    # Propagate to root if no user-defined module or a root handler is detected
    logger.propagate = False if _USER_MODULE_HANDLERS[logger.name] or not _USER_ROOT_HANDLER else True

    # Add or messages will repeat several times due to multiple handlers if same name used
    if not default_handlers and not (parallel_log_config or logger.name == parallel_module and top_level):
        # If no user specified default handler, any handler is assigned by OS and is cleared
        if logger.name == parallel_module:
            logger.handlers.clear()

        if flush:
            handler = _Flush(sys.stdout)
        else:
            handler = logging.StreamHandler(sys.stdout)

        handler.setFormatter(logging.Formatter("%(asctime)s %(name)s [%(levelname)s] %(message)s"))
        logger.addHandler(handler)

    return logger



import copy, os
from typing import Union, Optional

import numpy as np
from numpy.typing import NDArray

from .._utils import _convert_pickle_to_dict, _dicts_to_pickles, _logger

LG = _logger(__name__)


def change_dtype(
    subject_timeseries_list: Union[list[dict[str, dict[str, NDArray[np.floating]]]], list[os.PathLike]],
    dtype: Union[str, np.floating],
    return_dicts: bool = True,
    output_dir: Optional[os.PathLike] = None,
    filenames: Optional[list[str]] = None,
) -> dict[str, dict[str, dict[str, NDArray[np.floating]]]]:
    """
    Perform Participant-wise Dtype Conversion.

    Changes the dtypes of each participants NumPy array. This function uses the ``.astype()`` method from NumPy.
    This function can help reduce memory usage. For example, converting a NumPy array from "float64" to "float32" can
    halve the memory required, which is particularly useful when analyzing large datasets on a local machine.

    Parameters
    ----------
    subject_timeseries_list: :obj:`list[dict[str, dict[str, np.ndarray]]]` or :obj:`list[os.PathLike]`
        A list where each element consist of a dictionary mapping subject IDs to their run IDs and associated
        timeseries (TRs x ROIs) as a NumPy array. Can also be a list consisting of paths to pickle files
        containing this same structure. The expected structure of each dictionary is as follows:

        ::

            subject_timeseries = {
                    "101": {
                        "run-0": np.array([...]), # Shape: TRs x ROIs
                        "run-1": np.array([...]), # Shape: TRs x ROIs
                        "run-2": np.array([...]), # Shape: TRs x ROIs
                    },
                    "102": {
                        "run-0": np.array([...]), # Shape: TRs x ROIs
                        "run-1": np.array([...]), # Shape: TRs x ROIs
                    }
                }

    dtype: :obj:`bool` or :obj:`np.floating`
        Target data type (e.g "float32" or ``np.float32``) to convert each participant's NumPy arrays into.

    return_dicts: :obj:`bool`, default=True
        If True, returns a single dictionary containing the converted input dictionaries. Keys are named "dict_{0}"
        where {0} corresponds to the dictionary's position in the input list.

    output_dir: :obj:`os.PathLike` or :obj:`None`, default=None
        Directory to save the converted ``subject_timeseries`` as pickle files. The directory will be created if it
        does not exist. Dictionaries will not be saved if None.

    filenames: :obj:`list[str]` or :obj:`None`, default=None
        A list of names to save the dictionaries with changed dtypes as. Names are matched to dictionaries by position
        (e.g., a file name in the 0th position will be the file name for the dictionary in the 0th position of
        ``subject_timeseries_list``). If None and ``output_dir`` is specified, uses default file names -
        "subject_timeseries_{0}_float{1}.pkl" (where {0} indicates the original input order and {1} is the dtype.

        .. versionchanged:: 0.19.0  ``file_names`` to ``filenames``

    Returns
    -------
    dict[str, dict[str, dict[str, np.ndarray]]]
        A nested dictionary containing the converted subject timeseries.

    Warning
    -------
    **Floating Point Precision**: While this function allows conversion to any valid NumPy dtype, it is recommended to
    use floating-point dtypes. Reducing the dtype could introduce rounding errors that may lower the precision of
    subsequent analyses as decimal digits are reduced when lower dtypes are requested. Thus, the lowest recommended
    floating-point dtype would be "float32", since it allows for memory usage reduction while limiting rounding errors
    that could significantly  alter calculations.
    """
    assert isinstance(subject_timeseries_list, list), "`subject_timeseries_list` must be a list."

    if filenames is not None and output_dir is None:
        LG.warning("`filenames` supplied but no `output_dir` specified. Files will not be saved.")

    changed_dtype_dicts = {}

    for indx, curr_dict in enumerate(subject_timeseries_list):
        if isinstance(curr_dict, str) and curr_dict.endswith(".pkl"):
            curr_dict = _convert_pickle_to_dict(curr_dict)
        else:
            curr_dict = copy.deepcopy(curr_dict)

        for subj_id in curr_dict:
            for run in curr_dict[subj_id]:
                curr_dict[subj_id][run] = curr_dict[subj_id][run].astype(dtype)

        changed_dtype_dicts[f"dict_{indx}"] = curr_dict

    if output_dir:
        _dicts_to_pickles(
            output_dir=output_dir, dict_list=changed_dtype_dicts, filenames=filenames, call="change_dtype"
        )

    if return_dicts:
        return changed_dtype_dicts



"""A class which is responsible for accessing all CAP metadata and to keep track of all attributes in CAP"""

import copy, os
from typing import Union

import numpy as np
from ..check_parcel_approach import _check_parcel_approach
from sklearn.cluster import KMeans


class _CAPGetter:
    def __init__(self):
        pass

    ### Attributes exist when CAP initialized
    @property
    def parcel_approach(
        self,
    ) -> Union[
        dict[str, dict[str, Union[os.PathLike, list[str]]]],
        dict[str, dict[str, Union[os.PathLike, list[str], dict[str, dict[str, list[int]]]]]],
    ]:
        return self._parcel_approach

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict):
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    @property
    def groups(self) -> Union[dict[str, list[str]], None]:
        return self._groups

    ### Attributes exist when CAP.get_caps() used
    @property
    def n_clusters(self) -> Union[int, list[int], None]:
        return self._n_clusters if hasattr(self, "_n_clusters") else None

    @property
    def cluster_selection_method(self) -> Union[str, None]:
        return self._cluster_selection_method if hasattr(self, "_cluster_selection_method") else None

    @property
    def n_cores(self) -> Union[int, None]:
        return self._n_cores if hasattr(self, "_n_cores") else None

    @property
    def runs(self) -> Union[list[Union[int, str]], None]:
        return self._runs if hasattr(self, "_runs") else None

    @property
    def caps(self) -> Union[dict[str, dict[str, np.array]], None]:
        return self._caps if hasattr(self, "_caps") else None

    @property
    def kmeans(self) -> Union[dict[str, KMeans], None]:
        return self._kmeans if hasattr(self, "_kmeans") else None

    @property
    def cluster_scores(self) -> Union[dict[str, Union[str, dict[str, float]]], None]:
        return self._cluster_scores if hasattr(self, "_cluster_scores") else None

    @property
    def variance_explained(self) -> Union[dict[str, float], None]:
        return self._variance_explained if hasattr(self, "_variance_explained") else None

    @property
    def optimal_n_clusters(self) -> Union[dict[str, int], None]:
        return self._optimal_n_clusters if hasattr(self, "_optimal_n_clusters") else None

    @property
    def standardize(self) -> Union[bool, None]:
        return self._standardize if hasattr(self, "_standardize") else None

    @property
    def means(self) -> Union[dict[str, np.array], None]:
        return self._mean_vec if hasattr(self, "_mean_vec") else None

    @property
    def stdev(self) -> Union[dict[str, np.array], None]:
        return self._stdev_vec if hasattr(self, "_stdev_vec") else None

    @property
    def concatenated_timeseries(self) -> Union[dict[str, np.array], None]:
        return self._concatenated_timeseries if hasattr(self, "_concatenated_timeseries") else None

    @concatenated_timeseries.deleter
    def concatenated_timeseries(self):
        del self._concatenated_timeseries

    # Generated in `caps2plot`
    @property
    def region_caps(self) -> Union[dict[str, dict[str, np.array]], None]:
        return self._region_caps if hasattr(self, "_region_caps") else None

    @property
    def outer_products(self) -> Union[dict[str, dict[str, np.array]], None]:
        return self._outer_products if hasattr(self, "_outer_product") else None

    @property
    def subject_table(self) -> Union[dict[str, str], None]:
        return self._subject_table if hasattr(self, "_subject_table") else None

    @subject_table.setter
    def subject_table(self, subject_dict):
        if isinstance(subject_dict, dict):
            self._subject_table = copy.deepcopy(subject_dict)
        else:
            raise TypeError(
                "`self.subject_table` must be a dictionary where the keys are the subject IDs and the "
                "values are the group names."
            )

    @property
    def cosine_similarity(self) -> Union[dict[str, Union[list[str], dict[str, dict[str, float]]]], None]:
        return self._cosine_similarity if hasattr(self, "_cosine_similarity") else None



"""Internal function to extract timeseries with or without joblib"""

import copy, json, math, os, re
from dataclasses import dataclass, field
from functools import cached_property

import numpy as np, pandas as pd
from nilearn.maskers import NiftiLabelsMasker
from nilearn.image import index_img, load_img

from ..logger import _logger

# Logger initialization to check if any user-defined loggers where created prior to package import
LG = _logger(__name__)


# Data container
@dataclass
class _Data:
    parcel_approach: dict = field(default_factory=dict)
    signal_clean_info: dict = field(default_factory=dict)
    task_info: dict = field(default_factory=dict)
    tr: int = None
    verbose: bool = False

    # Run-specific attributes
    files: dict = field(default_factory=dict)
    fail_fast: bool = False
    dummy_vols: int = None
    censor_vols: list = field(default_factory=list)
    sample_mask: list = field(default_factory=list)
    max_len: int = None
    # Event condition scan indices
    scans: list = field(default_factory=list)
    # Subject header
    head: str = None
    # Percentage of volumes that exceed fd threshold
    vols_exceed_percent: float = None

    @property
    def session(self):
        return self.task_info["session"]

    @property
    def task(self):
        return self.task_info["task"]

    @property
    def condition(self):
        return self.task_info["condition"]

    @property
    def shift(self):
        return self.task_info["condition_tr_shift"]

    @property
    def slice_ref(self):
        return self.task_info["slice_time_ref"]

    @property
    def use_confounds(self):
        return self.signal_clean_info["use_confounds"]

    @property
    def fd(self):
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            return self.signal_clean_info["fd_threshold"]["threshold"]
        else:
            return self.signal_clean_info["fd_threshold"]

    @property
    def scrub_lim(self):
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            if "outlier_percentage" in self.signal_clean_info["fd_threshold"]:
                return self.signal_clean_info["fd_threshold"]["outlier_percentage"]
        else:
            return None

    @property
    def n_before(self):
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            if "n_before" in self.signal_clean_info["fd_threshold"]:
                return self.signal_clean_info["fd_threshold"]["n_before"]
        else:
            return None

    @property
    def n_after(self):
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            if "n_after" in self.signal_clean_info["fd_threshold"]:
                return self.signal_clean_info["fd_threshold"]["n_after"]
        else:
            return None

    @property
    def maps(self):
        return self.parcel_approach[list(self.parcel_approach)[0]]["maps"]

    @property
    def use_sample_mask(self):
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            if "use_sample_mask" in self.signal_clean_info["fd_threshold"]:
                return self.signal_clean_info["fd_threshold"]["use_sample_mask"]
            else:
                return False

    @property
    def censor_mask(self):
        return np.array(self.sample_mask) if len(self.sample_mask) > 0 else None

    @cached_property
    def confound_df(self):
        if self.files["confound"]:
            return pd.read_csv(self.files["confound"], sep="\t")
        else:
            return None


def _extract_timeseries(
    subj_id,
    prepped_files,
    run_list,
    parcel_approach,
    signal_clean_info,
    task_info,
    tr,
    verbose,
    flush,
    parallel_log_config=None,
):

    # Logger inside function to give logger to each child process if parallel processing is done
    LG = _logger(__name__, flush=flush, top_level=False, parallel_log_config=parallel_log_config)

    # Initialize subject dictionary
    subject_timeseries = {subj_id: {}}

    for run in run_list:
        # Initialize class; placing inside run loops allows re-initialization of defaults
        Data = _Data(parcel_approach, signal_clean_info, task_info, tr, verbose)

        # Due to prior checking in _setup_extraction within TimeseriesExtractor, run_list = [None] is assumed to be a
        # single file without the run- description
        run_id = "run-0" if run is None else run
        run = run if run is not None else ""

        # Get files from specific run
        Data.files = {
            "nifti": _grab(run, prepped_files["niftis"]),
            "mask": _grab(run, prepped_files["masks"]),
            "confound": _grab(run, prepped_files["confounds"]),
            "confound_meta": _grab(run, prepped_files["confounds_metas"]),
            "event": _grab(run, prepped_files["events"]),
        }

        # Base message
        Data.head = _header(Data, run_id.split("-")[-1], subj_id)

        if Data.verbose:
            LG.info(
                f"{Data.head}" + f"Preparing for Timeseries Extraction using "
                f"[FILE: {os.path.basename(Data.files['nifti'])}]."
            )

        # Get dummy volumes
        Data.dummy_vols = _get_dummy(Data, LG)

        # Assess framewise displacement
        if Data.fd and Data.files["confound"] and "framewise_displacement" in Data.confound_df.columns:
            # Get censor volumes vector, outlier_limit, and threshold
            Data.max_len, Data.censor_vols = _censor(Data)

            if Data.censor_vols and (Data.n_before or Data.n_after):
                Data.censor_vols = _extended_censor(Data)

            if len(Data.censor_vols) == Data.max_len:
                LG.warning(
                    f"{Data.head}" + "Timeseries Extraction Skipped: Timeseries will be empty due to "
                    f"all volumes exceeding a framewise displacement of {Data.fd}."
                )
                continue

            # Create sample mask
            if Data.censor_vols and Data.use_sample_mask:
                Data.sample_mask = _create_sample_mask(Data)

            # Check if run fails fast due to percentage volumes exceeding user-specified scrub_lim
            if Data.scrub_lim and not Data.condition:
                Data.vols_exceed_percent, Data.fail_fast = _flag(len(Data.censor_vols), Data.max_len, Data.scrub_lim)

        elif Data.fd and Data.files["confound"] and "framewise_displacement" not in Data.confound_df.columns:
            LG.warning(
                f"{Data.head}" + "`fd_threshold` specified but 'framewise_displacement' column not "
                "found in the confound tsv file. Removal of volumes after nuisance regression will not be "
                "but timeseries extraction will continue."
            )

        # Get events
        if Data.files["event"]:
            event_df = pd.read_csv(Data.files["event"], sep="\t")
            # Get specific timing information for specific condition
            condition_df = event_df[event_df["trial_type"] == Data.condition]

            if condition_df.empty:
                LG.warning(
                    f"{Data.head}" + f"[CONDITION: {Data.condition}] Timeseries Extraction Skipped: The "
                    "requested condition does not exist in the 'trial_type' column of the event file."
                )
                continue

            # Get condition indices
            Data.scans = _get_condition(Data, condition_df, LG)

            if Data.censor_vols:
                Data.scans, n = _filter_condition(Data, LG)
                if Data.scrub_lim:
                    Data.vols_exceed_percent, Data.fail_fast = _flag(n - len(Data.scans), n, Data.scrub_lim)

            if not Data.scans:
                LG.warning(
                    f"{Data.head}" + f"[CONDITION: {Data.condition}] Timeseries Extraction Skipped: Timeseries "
                    "will be empty when filtered to only include volumes from the requested condition. Possibly "
                    "due to the TRs corresponding to the condition being removed by `dummy_scans` or filtered "
                    "due to exceeding threshold for `fd_threshold`."
                )
                continue

        if Data.fail_fast:
            percent_msg = f"Percentage of volumes exceeding the threshold limit is {Data.vols_exceed_percent * 100}%"
            if Data.condition:
                percent_msg = f"{percent_msg} for [CONDITION: {Data.condition}]"

            LG.warning(
                f"{Data.head}" + f"Timeseries Extraction Skipped: Run flagged due to more than "
                f"{Data.scrub_lim * 100}% of the volumes exceeding the framewise displacement threshold of "
                f"{Data.fd}. {percent_msg}."
            )
            continue

        # Continue extraction if the continue keyword isn't hit when assessing the fd threshold or events
        timeseries = _continue_extraction(Data, LG)

        if timeseries.shape[0] == 0:
            LG.warning(
                f"{Data.head}" + f"Timeseries is empty and will not be appended to the "
                "`subject_timeseries` dictionary."
            )
        else:
            subject_timeseries[subj_id].update({run_id: timeseries})

    if not subject_timeseries[subj_id]:
        LG.warning(f"{Data.head.split(' | RUN:')[0]}] Timeseries Extraction Skipped: No runs were extracted.")
        subject_timeseries = None

    return subject_timeseries


def _continue_extraction(Data, LG):
    # Extract confound information of interest and ensure confound file does not contain NAs
    if Data.use_confounds:
        confound_names = copy.deepcopy(Data.signal_clean_info["confound_names"])
        # Extract first "n" numbers of specified WM and CSF components
        if Data.files["confound_meta"]:
            confound_names = _acompcor(
                confound_names, Data.files["confound_meta"], Data.signal_clean_info["n_acompcor_separate"]
            )

        # Get confounds
        confounds = _get_confounds(Data, confound_names, LG)

    # Create the masker for extracting time series; strategy="mean" is the default for NiftiLabelsMasker,
    # added to make it clear in codebase that the mean is the default strategy used for reducing regions
    masker = NiftiLabelsMasker(
        mask_img=Data.files["mask"],
        labels_img=Data.maps,
        resampling_target="data",
        strategy="mean",
        t_r=Data.tr,
        **Data.signal_clean_info["masker_init"],
        clean__extrapolate=False,
    )

    # Load and discard volumes if needed
    nifti_img = load_img(Data.files["nifti"], dtype=Data.signal_clean_info["dtype"])
    # Get length of timeseries and append scan indices to ensure no scans are out of bounds due to shift
    if (Data.condition and Data.shift > 0) and (max(Data.scans) > nifti_img.shape[-1] - 1):
        Data.scans = _check_indices(Data, nifti_img.shape[-1], LG, False)

    if Data.dummy_vols:
        nifti_img = index_img(nifti_img, slice(Data.dummy_vols, None))
        if Data.use_confounds and confounds is not None:
            confounds.drop(list(range(0, Data.dummy_vols)), axis=0, inplace=True)

    # Extract timeseries
    if Data.use_confounds and confounds is not None:
        timeseries = masker.fit_transform(nifti_img, confounds=confounds, sample_mask=Data.censor_mask)
    else:
        timeseries = masker.fit_transform(nifti_img, sample_mask=Data.censor_mask)

    # If sample mask was not used, delete censored volumes
    if not Data.condition:
        if Data.censor_vols and not Data.use_sample_mask:
            timeseries = np.delete(timeseries, Data.censor_vols, axis=0)

    if Data.condition:
        # Temporarily pad to extract the correct indices if sample mask was used
        if Data.censor_vols and Data.use_sample_mask:
            timeseries = _pad(timeseries, Data)
        # If any out of bound indices, remove them instead of getting indexing error.
        if max(Data.scans) > timeseries.shape[0] - 1:
            Data.scans = _check_indices(Data, timeseries.shape[0], LG)

        if Data.verbose:
            LG.info(Data.head + f"Nuisance regression completed; extracting [CONDITION: {Data.condition}].")

        # Extract condition
        timeseries = timeseries[Data.scans, :]

    return timeseries


# Grab files
def _grab(run, files):
    if files:
        return [file for file in files if run in os.path.basename(file)][0]
    else:
        return None


# Create subject header
def _header(Data, run_id, subj_id):
    if Data.session:
        sess_id = Data.session
    else:
        base_filename = os.path.basename(Data.files["nifti"])
        sess = re.search("ses-(\\S+?)[-_]", base_filename)[0][:-1] if "ses-" in base_filename else None
        sess_id = sess.split("-")[-1] if sess else None

    sub_head = f"[SUBJECT: {subj_id} | SESSION: {sess_id} | TASK: {Data.task} | RUN: {run_id}]"

    return f"{sub_head} "


# Get dummy scan number
def _get_dummy(Data, LG):
    info = Data.signal_clean_info["dummy_scans"]

    if isinstance(info, dict) and info["auto"] is True:
        n, flag = len([col for col in Data.confound_df.columns if "non_steady_state" in col]), "auto"

        if "min" in info and n < info["min"]:
            n, flag = info["min"], "min"
        if "max" in info and n > info["max"]:
            n, flag = info["max"], "max"
        if n == 0:
            n = None

        if Data.verbose:
            if flag == "auto":
                if n:
                    LG.info(
                        f"{Data.head}" + "Number of dummy scans to be removed based on "
                        f"'non_steady_state_outlier_XX' columns: {n}."
                    )
                else:
                    LG.info(
                        f"{Data.head}" + "No 'non_steady_state_outlier_XX' columns were found so 0 "
                        "dummy scans will be removed."
                    )
            else:
                LG.info(f"{Data.head}" + f"Default dummy scans set by '{flag}' will be used: {n}.")

    return n if isinstance(info, dict) else info


# Create censor vector
def _censor(Data):
    fd_array = Data.confound_df["framewise_displacement"].fillna(0).values

    # Truncate fd_array if dummy scans
    if Data.dummy_vols:
        fd_array = fd_array[Data.dummy_vols :]

    censor_volumes = sorted(list(np.where(fd_array > Data.fd)[0]))

    return fd_array.shape[0], censor_volumes


# Created an extended censor_vols vector
def _extended_censor(Data):
    ext_arr = []

    for i in Data.censor_vols:
        if Data.n_before:
            ext_arr.extend(list(range(i - Data.n_before, i)))
        if Data.n_after:
            ext_arr.extend(list(range(i + 1, i + Data.n_after + 1)))

    # Filter; ensure no index is below zero to prevent backwards indexing and not above max to prevent error
    filtered_ext_arr = [x for x in ext_arr if x >= 0 and x < Data.max_len]

    # Return new list that is sorted and only contains unique indices
    return sorted(list(set(Data.censor_vols + filtered_ext_arr)))


# Determine if run fails fast; when condition is not specified
def _flag(n_censor, n, scrub_lim):
    vols_exceed_percent = n_censor / n
    fail_fast = True if vols_exceed_percent > scrub_lim else False

    return vols_exceed_percent, fail_fast


# Get event condition
def _get_condition(Data, condition_df, LG):
    scans = []

    # Convert times into scan numbers to obtain the scans taken when the participant was exposed to the
    # condition of interest; include partial scans
    for i in condition_df.index:
        adjusted_onset = condition_df.loc[i, "onset"] - Data.slice_ref * Data.tr
        # Avoid accidental negative indexing
        adjusted_onset = adjusted_onset if adjusted_onset >= 0 else 0
        # Int is always the floor for positive floats
        onset_scan = int(adjusted_onset / Data.tr) + Data.shift
        end_scan = math.ceil((adjusted_onset + condition_df.loc[i, "duration"]) / Data.tr) + Data.shift
        # Add one since range is not inclusive
        scans.extend(list(range(onset_scan, end_scan + 1)))

    # Get unique scans to not duplicate information
    scans = sorted(list(set(scans)))

    # Adjust for dummy
    if Data.dummy_vols:
        scans = [scan - Data.dummy_vols for scan in scans if scan not in range(0, Data.dummy_vols)]

    return scans


def _filter_condition(Data, LG):
    # New `scans` list always created
    scans = Data.scans

    # Assess if any condition indices greater than fd array to not dilute outlier calculation
    if max(scans) > Data.max_len - 1:
        scans = _check_indices(Data, Data.max_len, LG)

    # Get length of scan list prior to assess outliers if requested
    n_before_censor = len(scans)
    scans = [volume for volume in scans if volume not in Data.censor_vols]

    return scans, n_before_censor


# Extract first "n" numbers of specified WM and CSF components
def _acompcor(confound_names, confound_metadata_file, n):
    with open(confound_metadata_file, "r") as confounds_json:
        confound_metadata = json.load(confounds_json)

    acompcors = sorted([acompcor for acompcor in confound_metadata if "a_comp_cor" in acompcor])

    CSF = [CSF for CSF in acompcors if confound_metadata[CSF]["Mask"] == "CSF"][0:n]
    WM = [WM for WM in acompcors if confound_metadata[WM]["Mask"] == "WM"][0:n]

    confound_names.extend(CSF + WM)

    return confound_names


# Get confounds
def _get_confounds(Data, confound_names, LG):
    valid_confounds = []
    invalid_confounds = []

    for confound_name in confound_names:
        if "*" in confound_name:
            prefix = confound_name.split("*")[0]
            confounds_list = [col for col in Data.confound_df.columns if col.startswith(prefix)]
        else:
            confounds_list = [col for col in Data.confound_df.columns if col == confound_name]

        if confounds_list:
            valid_confounds.extend(confounds_list)
        else:
            invalid_confounds.extend([confound_name])

    # First index of some variables is na
    if valid_confounds:
        confounds = Data.confound_df[valid_confounds].fillna(0)
    else:
        confounds = None

    if Data.verbose:
        if confounds is None:
            LG.warning(
                f"{Data.head}" + "None of the requested confounds were found so nuisance regression will not "
                "be done."
            )
        elif invalid_confounds:
            LG.warning(f"{Data.head}" + f"The following confounds were not found: {', '.join(invalid_confounds)}.")

        if confounds is not None and not confounds.empty:
            LG.info(
                f"{Data.head}" + "The following confounds will be used for nuisance regression: "
                f"{', '.join(list(confounds.columns))}."
            )

    return confounds


# Create sample mask for censoring high motion volumes for nuisance regression
def _create_sample_mask(Data):
    sample_mask = np.ones(Data.max_len, dtype="bool")
    sample_mask[np.array(Data.censor_vols)] = False

    return sample_mask


# Temporarily pad censored rows to correctly filter the condition
def _pad(timeseries, Data):
    padded_timeseries = np.zeros((Data.max_len, timeseries.shape[1]))
    padded_timeseries[Data.censor_mask, :] = timeseries

    return padded_timeseries


# Check if indices valid or truncate indices if shift applied and causes indices to be out of bounds
def _check_indices(Data, arr_shape, LG, warn=True):
    # Warn is false in instances were shift is used to prevent unnecessary logging, the warning is specifically
    # for instances related to misalignment such as potentially incorrect onsets and durations
    if warn:
        LG.warning(
            f"{Data.head}" + f"[CONDITION: {Data.condition}] Max scan index exceeds timeseries max index. "
            f"Max condition index is {max(Data.scans)}, while max timeseries index is {arr_shape - 1}. Timing may "
            "be misaligned or specified repetition time incorrect. If intentional, ignore warning. Only "
            "indices for condition within the timeseries range will be extracted."
        )

    return [scan for scan in Data.scans if scan in range(0, arr_shape)]



"""A class which is responsible for accessing all TimeseriesExtractorGetter and to keep track of all attributes in
TimeSeriesExtractor"""

import copy, os
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..check_parcel_approach import _check_parcel_approach
from ..pickle_utils import _convert_pickle_to_dict


class _TimeseriesExtractorGetter:
    def __init__(self):
        pass

    #### Exists upon initialization of TimeseriesExtractor
    @property
    def space(self) -> str:
        return self._space

    @space.setter
    def space(self, new_space):
        if not isinstance(new_space, str):
            raise TypeError("`space` must be a string.")
        self._space = new_space

    @property
    def parcel_approach(
        self,
    ) -> Union[
        dict[str, dict[str, Union[os.PathLike, list[str]]]],
        dict[str, dict[str, Union[os.PathLike, list[str], dict[str, dict[str, list[int]]]]]],
    ]:
        return self._parcel_approach

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict):
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    @property
    def signal_clean_info(self) -> Union[dict[str, Union[bool, int, float, str]], None]:
        return self._signal_clean_info

    ### Does not exists upon initialization of Timeseries Extractor

    # Exist when TimeSeriesExtractor.get_bold() used
    @property
    def task_info(self) -> Union[dict[str, Union[str, int]], None]:
        return self._task_info if hasattr(self, "_task_info") else None

    # Gets initialized and populated in TimeSeriesExtractor.get_bold(),
    @property
    def subject_ids(self) -> Union[list[str], None]:
        return self._subject_ids if hasattr(self, "_subject_ids") else None

    @property
    def n_cores(self) -> Union[int, None]:
        return self._n_cores if hasattr(self, "_n_cores") else None

    # Gets initialized in TimeSeriesExtractor.get_bold(), gets populated when
    # TimeseriesExtractor._timeseries_aggregator gets called in TimeseriesExtractor._extract_timeseries
    @property
    def subject_timeseries(self) -> Union[dict[str, dict[str, NDArray[np.floating]]], None]:
        return self._subject_timeseries if hasattr(self, "_subject_timeseries") else None

    @subject_timeseries.setter
    def subject_timeseries(self, subject_dict):
        need_deepcopy = True

        if isinstance(subject_dict, str) and subject_dict.endswith(".pkl"):
            subject_dict = _convert_pickle_to_dict(subject_dict)
            need_deepcopy = False

        self._validate_timeseries(subject_dict)

        if need_deepcopy:
            self._subject_timeseries = copy.deepcopy(subject_dict)
        else:
            self._subject_timeseries = subject_dict

    @subject_timeseries.deleter
    def subject_timeseries(self):
        del self._subject_timeseries

    @staticmethod
    def _validate_timeseries(subject_dict):
        error_msg = (
            "A valid pickle file/subject timeseries should contain a nested dictionary where the "
            "first level is the subject id, second level is the run number in the form of 'run-#', and "
            "the final level is the timeseries as a numpy array. "
        )

        error_dict = {
            "Sub": error_msg + "The error occurred at [SUBJECT: {0}]. ",
            "Run": error_msg + "The error occurred at [SUBJECT: {0} | RUN: {1}]. ",
        }

        if not isinstance(subject_dict, dict):
            raise TypeError(error_msg)

        for sub in subject_dict:
            if not isinstance(subject_dict[sub], dict):
                raise TypeError(
                    error_dict["Sub"].format(sub) + "The subject must be a dictionary with second level "
                    "'run-#' keys."
                )

            runs = list(subject_dict[sub])

            if not all("run" in x for x in runs):
                raise TypeError(error_dict["Sub"].format(sub) + "Not all second level keys follow the form of 'run-#'.")

            for run in runs:
                if not isinstance(subject_dict[sub][run], np.ndarray):
                    raise TypeError(error_dict["Run"].format(sub, run) + "All 'run-#' keys must contain a numpy array.")
